#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 22:08

""" pyxllib常用功能
"""

from pyxllib.stdlib import zipfile
from pyxllib.stdlib import shutil

from pyxllib.prog.newbie import *
from pyxllib.prog.pupil import *
from pyxllib.prog.specialist import *
from pyxllib.prog.deprecatedlib import deprecated

from pyxllib.algo.newbie import *
from pyxllib.algo.pupil import *
from pyxllib.algo.specialist import *

from pyxllib.text.newbie import *
from pyxllib.text.pupil import *
from pyxllib.text.specialist import *

from pyxllib.file.newbie import *
from pyxllib.file.pupil import *
from pyxllib.file.specialist import *

from pyxllib.debug.newbie import *
from pyxllib.debug.pupil import *
from pyxllib.debug.specialist import *
